#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "clock.h"
#include "wav.h"


/* Filters */
void lowpass( wav_file *wav ) {
  uint64_t clockStart = x86clock();
  const float coefficients[4] = { 0.378f, 0.623f, 0.623f, 0.378f };
  float samples[4] = {
    wav->samples[0], wav->samples[0], wav->samples[1], wav->samples[2] 
  };
  
#ifdef USE_SSE
  float result[4];
#endif

  int i;
  for( i = 3; i < wav->numsamples; i++ ) {

    /* Neuen Wert einlesen */
    samples[0] = samples[1];
    samples[1] = samples[2];
    samples[2] = samples[3];
    samples[3] = wav->samples[i];

#ifdef USE_SSE

    // TODO: Hier soll die Implementierung mit SSE-Instruktionen stehen

#else

    // TODO: Hier soll eine Implementierung in Standard-C stehen
    //       (ohne Intrinsics und Inline Assembly)
    // 
    //       Hinweis: Für beide Implementierungen können Sie obigen 
    //       Koeffizienten-Array nutzen.

#endif
  }
  
  uint64_t clockEnd = x86clock();
  printf( "Time taken: %lli\n", clockEnd - clockStart );
}

void noise( wav_file *wav, int thresh ) {
  int i, a, b;
  for( i = 0; i < wav->numsamples; i++ ) {
    if( abs( wav->samples[i] ) < thresh ) {
      a = rand() % thresh;
      b = rand() % 1024;
      if( b & 1 )
        a = -a;
      wav->samples[i] += a;
    }
  }
}

void echo( wav_file *wav, unsigned int Delay, unsigned int NumEcho ) {
  uint64_t clockStart = x86clock();
  
  int i, k;
  
  // Start echoing
#ifdef USE_MMX

  // TODO: Hier soll eine MMX-Implementierung der entrollten 
  //       Echo-Schleife stehen. Für MMX ist es günstig, die
  //       Schleife um Faktor 4 abzurollen (1 Iteration verarbeitet
  //       ein 16-Bit Wort, und die MMX-Register können 64-Bit
  //       fassen, d.h. MMX-Operationen können 4 16-Bit-Werte
  //       gleichzeitig bearbeiten).
  //       Beachten Sie dabei, daß die Schleifeniterationszahl
  //       nicht zwangsläufig durch 4 teilbar sein muß. Lassen
  //       sie daher in einer vorgeschalteten Schleife solange
  //       Iterationen ausführen, bis die Anzahl der verbleibenden 
  //       Iterationen durch 4 teilbar ist.
  //       Danach können Sie die zweite Schleife in 4er-Schritten
  //       durch das Array laufen lassen.

#else
  for( k = 1; k <= NumEcho; k++ ) {
    for( i = wav->numsamples - 1; i >= Delay*k; i-- ) {
      wav->samples[i] += wav->samples[i - Delay * k] >> k;
    }
  }
#endif
  
  uint64_t clockEnd = x86clock();
  printf( "Time taken: %lli\n", clockEnd - clockStart );
}

/* Main Program */
int main(int argc,char **argv)
{
  // Check arguments
  if (argc<3) {
    printf("Usage: %s <input-wav filename> <filter-name> <output-wav filename>\n",argv[0]);
    return(1);
  }
  char *inputFile = argv[1];
  char *outputFile = argv[3];
  char *filter = argv[2];

  // Read WAV  
  wav_file *inputWAV = readWAV( inputFile );

  // Filter
  if ( strcmp( filter, "echo" ) == 0 ) {
    echo( inputWAV, 500, 10 );
  } else
  if ( strcmp( filter, "lowpass" ) == 0 ) {
    lowpass( inputWAV );
  } else
  if ( strcmp( filter, "noise" ) == 0 ) {
    noise( inputWAV, 750 );
  } else {
    printf("Unknown filter '%s'!\n", filter);
    exit(3);
  }

  // Write result
  writeWAV( inputWAV, outputFile );

  // Free the WAV struct
  freeWAV( inputWAV );

  return 0;
}
